/*

  File: Piece.cpp

  (c) 2003 by Oliver Ehli

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

     See File LICENSE for more Information.

*/
#include <string.h>
#include <fstream.h>
#include "Piece.h"
#include "InventorWriter.h"


long long Piece::mask = (1ll << PC_X_DIM) - 1ll;

Piece::Piece(int inId, int (*dimensions)[3], bool (*description)[PC_Z_DIM][PC_Y_DIM][PC_X_DIM],
	     bool skipSymmetry, bool specialSymmetry) : numPositions(0), numRotations(1)
{
  _bbox = dimensions;
  _cubes = description;

  id = inId;

  piece = 0;

  for (int i=0; i<GC_X_DIM; i++)
    for (int j=0; j<GC_Y_DIM; j++)
      for (int k=0; k<GC_Z_DIM; k++) {
	if(i>=(*_bbox)[2] || j>=(*_bbox)[1] || k>=(*_bbox)[0]) continue;
	if((*_cubes)[i][j][k])
	  piece |= ((1ll << ((k * GC_Y_DIM + j) * GC_X_DIM)) << i);
      }

  int tmp = (*_bbox)[0];
  (*_bbox)[0] = (*_bbox)[2];
  (*_bbox)[2] = tmp;

  rotations[0] = piece;
  rotatedBBoxes = new int*[MAX_NUM_ROT];
  rotatedBBoxes[0] = *dimensions;

  buildRotations();
  buildPositions(skipSymmetry, specialSymmetry);

  for (int i=0; i<GC_X_DIM*GC_Y_DIM*GC_Z_DIM; i++)
    occupiedCoordsIndex[i] = 0;
  buildOuccupanceInfo();

}


long long Piece::addRotation(long long rotatedPiece, int* rotatedBBox)
{
  int i;
  for (i=0; i< numRotations; i++)
    if (rotations[i] == rotatedPiece) break;
  
  if(i==numRotations) {
    rotations[i] = rotatedPiece;
    rotatedBBoxes[i] = rotatedBBox;
    numRotations++;
  }

  return rotatedPiece;
}


void Piece::buildRotations()
{
  long long current  = piece;

  for(int i=1; i<4; i++) {
    int *tmpXBBox = new int[3];

    tmpXBBox[0] = (*_bbox)[0];
    tmpXBBox[1] = (*_bbox)[2];
    tmpXBBox[2] = (*_bbox)[1];

    if((i%2 && tmpXBBox[1]>GC_Y_DIM) || (i%2 && tmpXBBox[2]>GC_Z_DIM)) continue;
    addRotation(Puzzle::rotate(current, i, 0, 0, *_bbox, mask), (i % 2 ? tmpXBBox : *_bbox));
  }

  int nr = numRotations;
  for(int i=0; i<nr; i++) {
    int *tmpYBBox = new int[3];
    tmpYBBox[0] = rotatedBBoxes[i][2];
    tmpYBBox[1] = rotatedBBoxes[i][1];
    tmpYBBox[2] = rotatedBBoxes[i][0];

    for(int j=1; j<4; j++) {
      if((j%2 && tmpYBBox[0]>GC_X_DIM) || (j%2 && tmpYBBox[2]>GC_Z_DIM)) continue;
      addRotation(Puzzle::rotate(rotations[i], 0, j, 0, rotatedBBoxes[i], mask),
		  (j % 2 ? tmpYBBox : rotatedBBoxes[i]));
    }
  }

  nr = numRotations;
  for(int i=0; i<nr; i++) {
    int *tmpZBBox = new int[3];
    tmpZBBox[0] = rotatedBBoxes[i][1];
    tmpZBBox[1] = rotatedBBoxes[i][0];
    tmpZBBox[2] = rotatedBBoxes[i][2];

    for(int j=1; j<4; j++) {
      if((j%2 && tmpZBBox[0]>GC_X_DIM) || (j%2 && tmpZBBox[1]>GC_Y_DIM)) continue;
      addRotation(Puzzle::rotate(rotations[i], 0, 0, j, rotatedBBoxes[i], mask),
		  (j % 2 ? tmpZBBox : rotatedBBoxes[i]));
    }
  }

}

long long Piece::mirror(long long piece, bool xAxis, bool yAxis, bool zAxis)
{
  long long result = 0;
  long long mask = 0;

  for(int i=0; i<GC_Z_DIM; i++) 
    for(int j=0; j<GC_Y_DIM; j++) {
      mask = ((long long) 1) << ((i*GC_Y_DIM + j) * GC_X_DIM);
      for(int k=0; k<GC_X_DIM; k++, mask <<= 1)
	if(piece & mask) {
	  result |= ((((long long) 1) <<
		      (xAxis ? GC_X_DIM-1-k : k)) <<
		     (((zAxis ? GC_Z_DIM-1-i : i)*GC_Y_DIM +
		       (yAxis ? GC_Y_DIM-1-j : j)) * GC_X_DIM));
	}
    }

  return result;
}


void Piece::buildPositions(bool skipSymmetry, bool specialSymmetry)
{

  if(!specialSymmetry)
    for(int l=0; l < GC_Z_DIM; l++) 
      for(int k=0; k < GC_Y_DIM; k++) {
	int trY = (l*GC_Y_DIM + k) * GC_X_DIM;
	for(int j=0; j < GC_X_DIM; j++)
	  for(int i=0; i< numRotations; i++) {
	    if (j > GC_X_DIM - rotatedBBoxes[i][0] ||
		k > GC_Y_DIM - rotatedBBoxes[i][1] ||
		l > GC_Z_DIM - rotatedBBoxes[i][2]) continue;
	    long long nextPiece = ((rotations[i] <<j) << trY);
	    if(skipSymmetry) {
	      bool found = false;
	      for(int pos=0; pos < numPositions; pos++)
		if(positions[pos] == mirror(nextPiece, true, true, false) ||
		   positions[pos] == mirror(nextPiece, true, false, true) ||
		   positions[pos] == mirror(nextPiece, false, true, true) ||
		   positions[pos] == nextPiece)
		  found = true;
	      if(found) continue;
	    }
	    positions[numPositions] = nextPiece;
	    positionRotation[numPositions] = i;
	    numPositions++;
	  }
      }

  else
    for(int l=GC_Z_DIM-1; l>=0; l--) 
      for(int k=GC_Y_DIM-1; k>=0; k--) {
	int trY = (l*GC_Y_DIM + k) * GC_X_DIM;
	for(int j=GC_X_DIM-1; j>=0; j--)
	  for(int i=0; i< numRotations; i++) {
	    if (j > GC_X_DIM - rotatedBBoxes[i][0] ||
		k > GC_Y_DIM - rotatedBBoxes[i][1] ||
		l > GC_Z_DIM - rotatedBBoxes[i][2]) continue;
	    long long nextPiece = ((rotations[i] <<j) << trY);
	    {
	      bool found = false;
	      for(int pos=0; pos < numPositions; pos++)
		if(positions[pos] == mirror(nextPiece, true, false, true))
		  found = true;
	      if(found) continue;
	    }
	    positions[numPositions] = nextPiece;
	    positionRotation[numPositions] = i;
	    numPositions++;
	  }
      }
}
 



void Piece::buildOuccupanceInfo()
{

  long long mask = 1;
  int maxPos = 0;

  for(int j=0; j < GC_Z_DIM; j++)
    for(int k=0; k < GC_Y_DIM; k++)
      for(int l=0; l < GC_X_DIM; l++, mask <<= 1)
	for(int i=0; i<numPositions; i++)
	  if(positions[i] & mask && !(positions[i] & (mask-1))) {

	    int off = (j*GC_Y_DIM + k)*GC_X_DIM + l + 4;
	    occupiedCoords[off][occupiedCoordsIndex[off]] = positions[i] << 4;
	    occupiedCoordsPositionIndex[off][occupiedCoordsIndex[off]] = i;
	    occupiedCoordsIndex[off]++;
	  }

}

void Piece::dumpPositionInfo()
{
  cout << "int piece" << id << "NumPositions = " << numPositions << ";" << endl;

  cout << "long long piece" << id << "Positions[" << numPositions << "] = {" << endl << "  ";
  for (int i=0; i < numPositions; i++) {
    cout << "0x" << hex << positions[i] << ", ";
      if(!((i+1)%4)) cout << endl << "  ";
  }
  cout << dec << endl << "};" << endl;
}




void Piece::dumpOuccupanceInfo(bool hi)
{

  int start = hi ? 32 : 0;
  int stop = hi ? GC_X_DIM*GC_Y_DIM*GC_Z_DIM : 32;

  cout << "   {" << endl;
  for (int i=start; i < stop; i++) {
    cout << "\t{  ";
    for (int j=0; j < occupiedCoordsIndex[i]; j++) {
      if(hi)
	cout << "0x" << hex << (int)(occupiedCoords[i][j]>>32) << ", ";
      else
	cout << "0x" << hex << occupiedCoords[i][j] << ", ";
      if(!((j+1)%4)) cout << endl << "\t   ";
    }
    cout << endl << "\t}," << endl;
  }

  cout << dec << "   }," << endl << endl << endl;

}




void Piece::dumpOuccupanceIndexInfo(bool hi)
{
  int start = hi ? 32 : 0;
  int stop = hi ? GC_X_DIM*GC_Y_DIM*GC_Z_DIM : 32;

  cout << "   {" << endl;

  for (int i=start; i < stop; i++) {
    if(!((i-start)%10)) cout << "     ";
    cout << occupiedCoordsIndex[i] << ", ";
    if(!((i-start+1)%10)) cout << endl;
  }
  cout << dec << endl << "   }," << endl << endl;
}




// debug-stuff below

#ifdef DEBUG_BIN

void Piece::writeRotatedPiecesAsIV(ofstream & out, const int indentLevel, const int pieceNo)
{

  for(int l=0; l<numRotations; l++) {
    out << InventorWriter::indent;
    out << InventorWriter::beginGroup;

    InventorWriter::writeTranslation(out,
				     l%5 *  7 * InventorWriter::baseUnit,
				     l/5 * -5 * InventorWriter::baseUnit, 0.0);

    InventorWriter::writeBBox(out, indentLevel+1, rotatedBBoxes[l]);

    int m = 0, hiLo = 0;
    for (int i=0; i<GC_Z_DIM; i++) {
      for (int j=0; j<GC_Y_DIM; j++)
	for (int k=0; k<GC_X_DIM; k++, m++)
	  if(!(rotations[l] & (1ll << m))) continue;
	  else writeAsIV(out, indentLevel+1, pieceNo, k, j, i);
    }

    out << InventorWriter::indent;
    out << InventorWriter::endGroup;
  }

}



void Piece::writePiecePositionsAsIV(ofstream & out, const int indentLevel, const int pieceNo)
{

  int bbox[] = { 3, 4, 5 };

  for(int l=0; l<numPositions; l++) {
    out << InventorWriter::indent;
    out << InventorWriter::beginGroup;

    InventorWriter::writeTranslation(out,
				     l%5 *  7 * InventorWriter::baseUnit,
				     l/5 * -5 * InventorWriter::baseUnit, 0.0);

    InventorWriter::writeBBox(out, indentLevel+1, bbox);

    int m = 0, hiLo = 0;
    for (int i=0; i<GC_Z_DIM; i++) {
      for (int j=0; j<GC_Y_DIM; j++)
	for (int k=0; k<GC_X_DIM; k++, m++) {
	  if((positions[l] & (1ll << m)))
	    writeAsIV(out, indentLevel+1, pieceNo, k, j, i);
	}
    }

    out << InventorWriter::indent;
    out << InventorWriter::endGroup;
  }

}



void Piece::writeDescriptionAsIV(ofstream & out, const int indentLevel, const int pieceNo)
{
  for (int i=0; i<(*_bbox)[2]; i++)
    for (int j=0; j<(*_bbox)[1]; j++)
      for (int k=0; k<(*_bbox)[0]; k++)
	if(!(*_cubes)[i][j][k]) continue;
	else writeAsIV(out, indentLevel, pieceNo, i, j, k);
}


void Piece::writePieceAsIV(ofstream & out, const int indentLevel, const int pieceNo)
{
  int m = 0, hiLo = 0;
  for (int i=0; i<GC_Z_DIM; i++) {
    for (int j=0; j<GC_Y_DIM; j++)
      for (int k=0; k<GC_X_DIM; k++, m++)
	if(!(piece & (1ll << m))) continue;
	else writeAsIV(out, indentLevel, pieceNo, i, j, k);
  }
}


void Piece::writeAsIV(ofstream & out, const int indentLevel, const int pieceNo,
		      const int x, const int y, const int z)
{

  for (int l=0; l<indentLevel; l++)
    out << InventorWriter::indent;
  out << InventorWriter::beginGroup;

  for (int l=0; l<=indentLevel; l++)
    out << InventorWriter::indent;
    InventorWriter::writeColor(out, pieceNo);

  for (int l=0; l<=indentLevel; l++)
    out << InventorWriter::indent;
  InventorWriter::writeTranslation(out,
				   x * InventorWriter::baseUnit,
				   y * InventorWriter::baseUnit,
				   z * InventorWriter::baseUnit);

  for (int l=0; l<=indentLevel; l++)
    out << InventorWriter::indent;
  InventorWriter::writeCube(out);

  for (int l=0; l<indentLevel; l++)
    out << InventorWriter::indent;
  out << InventorWriter::endGroup;

}

#endif
